"""Module to unroll HEBGraph.

Unrolling means expanding each sub-behavior node as it's own graph in the global HEBGraph.
Behaviors that do not have a graph (Unexplainable behaviors) should stay as is in the graph.

"""
from copy import copy
from typing import TYPE_CHECKING, Dict, List, Tuple, Optional, Union

from networkx import relabel_nodes

from hebg.behavior import Behavior

BEHAVIOR_SEPARATOR = ">"

if TYPE_CHECKING:
    from hebg import HEBGraph, Node, Action


def unroll_graph(
    graph: "HEBGraph",
    add_prefix=False,
    cut_looping_alternatives: bool = False,
) -> "HEBGraph":
    """Build the the unrolled HEBGraph.

    The HEBGraph as the same behavior but every behavior node is recursively replaced
    by it's own HEBGraph if it can be computed.

    Args:
        graph (HEBGraph): HEBGraph to unroll the behavior in.
        add_prefix (bool, optional): If True, adds a name prefix to keep nodes different.
            Defaults to False.
        cut_looping_alternatives (bool, optional): If True, cut the looping alternatives.
            Defaults to False.

    Returns:
        HEBGraph: This HEBGraph's unrolled HEBGraph.
    """
    unrolled_graph, _is_looping = _unroll_graph(
        graph,
        add_prefix=add_prefix,
        cut_looping_alternatives=cut_looping_alternatives,
    )
    return unrolled_graph


def _unroll_graph(
    graph: "HEBGraph",
    add_prefix=False,
    cut_looping_alternatives: bool = False,
    _current_alternatives: Optional[List[Union["Action", "Behavior"]]] = None,
    _unrolled_behaviors: Optional[Dict[str, Optional["HEBGraph"]]] = None,
) -> Tuple["HEBGraph", bool]:
    if _unrolled_behaviors is None:
        _unrolled_behaviors = {}
    if _current_alternatives is None:
        _current_alternatives = []

    is_looping = False
    _unrolled_behaviors[graph.behavior.name] = None

    unrolled_graph: "HEBGraph" = copy(graph)
    for node in list(unrolled_graph.nodes()):
        if not isinstance(node, Behavior):
            continue
        new_alternatives = []
        for pred, _node, data in graph.in_edges(node, data=True):
            index = data["index"]
            for _pred, alternative, alt_index in graph.out_edges(pred, data="index"):
                if index == alt_index and alternative != node:
                    new_alternatives.append(alternative)
        if new_alternatives:
            _current_alternatives = new_alternatives
        unrolled_graph, behavior_is_looping = _unroll_behavior(
            unrolled_graph,
            node,
            add_prefix,
            cut_looping_alternatives,
            _current_alternatives,
            _unrolled_behaviors,
        )
        if behavior_is_looping:
            is_looping = True

    return unrolled_graph, is_looping


def _unroll_behavior(
    graph: "HEBGraph",
    behavior: "Behavior",
    add_prefix: bool,
    cut_looping_alternatives: bool,
    _current_alternatives: List[Union["Action", "Behavior"]],
    _unrolled_behaviors: Dict[str, Optional["HEBGraph"]],
) -> Tuple["HEBGraph", bool]:
    """Unroll a behavior node in a given HEBGraph

    Args:
        graph (HEBGraph): HEBGraph to unroll the behavior in.
        behavior (Behavior): Behavior node to unroll, must be in the given graph.
        add_prefix (bool): If True, adds a name prefix to keep nodes different.
        cut_looping_alternatives (bool): If True, cut the looping alternatives.

    Returns:
        HEBGraph: Initial graph with unrolled behavior.
    """
    # Look for name reference.
    if behavior.name in graph.all_behaviors:
        behavior = graph.all_behaviors[behavior.name]

    node_graph, is_looping = _unrolled_behavior_graph(
        behavior,
        add_prefix,
        cut_looping_alternatives,
        _current_alternatives,
        _unrolled_behaviors,
    )

    if is_looping and cut_looping_alternatives:
        if not _current_alternatives:
            return graph, is_looping
        for alternative in _current_alternatives:
            for last_condition, _, data in graph.in_edges(behavior, data=True):
                graph.add_edge(last_condition, alternative, **data)
        graph.remove_node(behavior)
        return graph, False

    if node_graph is None:
        # If we cannot get the node's graph, we keep it as is.
        return graph, is_looping

    # Relabel graph nodes to obtain disjoint node labels (if more that one node).
    if add_prefix and len(node_graph.nodes()) > 1:
        _add_prefix_to_graph(node_graph, behavior.name + BEHAVIOR_SEPARATOR)

    # Replace the behavior node by the unrolled behavior's graph
    graph = compose_heb_graphs(graph, node_graph)
    for edge_u, _, data in graph.in_edges(behavior, data=True):
        for root in node_graph.roots:
            graph.add_edge(edge_u, root, **data)

    graph.remove_node(behavior)
    return graph, is_looping


def _unrolled_behavior_graph(
    behavior: "Behavior",
    add_prefix: bool,
    cut_looping_alternatives: bool,
    _current_alternatives: List[Union["Action", "Behavior"]],
    _unrolled_behaviors: Dict[str, Optional["HEBGraph"]],
) -> Optional["HEBGraph"]:
    """Get the unrolled sub-graph of a behavior.

    Args:
        behavior (Behavior): Behavior to get the unrolled graph of.
        add_prefix (bool): If True, adds a prefix in sub-hierarchies to have distinct nodes.
        cut_looping_alternatives (bool): If True, cut the looping alternatives.
        _unrolled_behaviors (Dict[str, Optional[HEBGraph]]): Dictionary of already computed
            unrolled graphs, both to save compute and prevent recursion loops.

    Returns:
        Optional[HEBGraph]: Unrolled graph of a behavior, None if it cannot be computed.
    """
    if behavior.name in _unrolled_behaviors:
        # If we have aleardy unrolled this behavior, we reuse it's graph
        is_looping = _unrolled_behaviors[behavior.name] is None
        return _unrolled_behaviors[behavior.name], is_looping

    try:
        node_graph, is_looping = _unroll_graph(
            behavior.graph,
            add_prefix=add_prefix,
            cut_looping_alternatives=cut_looping_alternatives,
            _current_alternatives=_current_alternatives,
            _unrolled_behaviors=_unrolled_behaviors,
        )
        _unrolled_behaviors[behavior.name] = node_graph
        return node_graph, is_looping
    except NotImplementedError:
        return None, False


def _add_prefix_to_graph(graph: "HEBGraph", prefix: str) -> None:
    """Rename graph to obtain disjoint node labels."""
    if prefix is None:
        return graph

    def rename(node: "Node"):
        new_node = copy(node)
        new_node.name = prefix + node.name
        return new_node

    return relabel_nodes(graph, rename, copy=False)


def group_behaviors_points(
    pos: Dict["Node", tuple],
    graph: "HEBGraph",
) -> Dict[tuple, list]:
    """Group nodes positions of an HEBGraph by sub-behavior.

    Args:
        pos (Dict[Node, tuple]): Positions of nodes.
        graph (HEBGraph): Graph.

    Returns:
        Dict[tuple, list]: A dictionary of nodes grouped by their behavior's hierarchy.
    """
    points_grouped_by_behavior: Dict[tuple, list] = {}
    for node in graph.nodes():
        groups = str(node).split(BEHAVIOR_SEPARATOR)
        if len(groups) > 1:
            for i in range(len(groups[:-1])):
                key = tuple(groups[: -1 - i])
                point = pos[node]
                try:
                    points_grouped_by_behavior[key].append(point)
                except KeyError:
                    points_grouped_by_behavior[key] = [point]
    return points_grouped_by_behavior


def compose_heb_graphs(graph_of_reference: "HEBGraph", other_graph: "HEBGraph"):
    """Returns a new_graph of graph_of_reference composed with other_graph.

    Composition is the simple union of the node sets and edge sets.
    The node sets of the graph_of_reference and other_graph do not need to be disjoint.

    Args:
        graph_of_reference, other_graph : HEBGraphs to compose.

    Returns:
        A new HEBGraph with the same type as graph_of_reference.

    """
    new_graph = graph_of_reference.__class__(
        graph_of_reference.behavior,
        all_behaviors=graph_of_reference.all_behaviors,
        any_mode=graph_of_reference.any_mode,
    )
    # add graph attributes, H attributes take precedent over G attributes
    new_graph.graph.update(graph_of_reference.graph)
    new_graph.graph.update(other_graph.graph)

    new_graph.add_nodes_from(graph_of_reference.nodes(data=True))
    new_graph.add_nodes_from(other_graph.nodes(data=True))

    if graph_of_reference.is_multigraph():
        new_graph.add_edges_from(graph_of_reference.edges(keys=True, data=True))
    else:
        new_graph.add_edges_from(graph_of_reference.edges(data=True))
    if other_graph.is_multigraph():
        new_graph.add_edges_from(other_graph.edges(keys=True, data=True))
    else:
        new_graph.add_edges_from(other_graph.edges(data=True))
    return new_graph



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Simplest binary utility for HEBGraph. """

from typing import Dict, List

from hebg import Behavior


def binary_graphbased_utility(
    behavior: Behavior,
    solving_behaviors: List[Behavior],
    used_nodes: Dict[str, Dict[str, int]],
) -> bool:
    """Returns if the behavior is in the HEBGraph of any solving_behavior.

    Args:
        behavior: Behavior of which we want to compute the utility.
        solving_behaviors: list of behaviors that solves the task of interest.
        used_nodes: dictionary mapping behavior_id to nodes used in the behavior.

    Returns:
        True if the behavior in the HEBGraph of any solving_behavior. False otherwise.

    """

    for solving_behavior in solving_behaviors:
        if behavior == solving_behavior:
            return True
        if behavior in solving_behavior.graph.nodes():
            return True
        if (
            behavior in used_nodes[solving_behavior]
            and used_nodes[solving_behavior][behavior] > 0
        ):
            return True
    return False



from typing import Dict

from hebg.node import Action, FeatureCondition
from hebg.behavior import Behavior
from hebg.heb_graph import HEBGraph

from tests.examples.feature_conditions import ThresholdFeatureCondition


class FundamentalBehavior(Behavior):

    """Fundamental behavior based on an Action."""

    def __init__(self, action: Action) -> None:
        self.action = action
        name = action.name + "_behavior"
        super().__init__(name, image=action.image)

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        graph.add_node(self.action)
        return graph


class AA_Behavior(Behavior):

    """Double root fundamental behavior"""

    def __init__(self, name: str, any_mode: str) -> None:
        super().__init__(name, image=None)
        self.any_mode = any_mode

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self, any_mode=self.any_mode)

        graph.add_node(Action(0))
        graph.add_node(Action(1))

        return graph


class F_A_Behavior(Behavior):

    """Single feature condition behavior"""

    def __init__(
        self,
        name: str,
        feature_condition: FeatureCondition,
        actions: Dict[int, Action],
    ) -> None:
        """Single feature condition behavior

        Args:
            name (str): Name of the behavior.
            feature_condition (FeatureCondition): Feature_condition used in behavior.
            actions (Dict[int, Action]): Mapping from feature_condition output to actions.
        """
        super().__init__(name)
        self.actions = actions
        self.feature_condition = feature_condition

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        for fc_output, action in self.actions.items():
            graph.add_edge(self.feature_condition, action, index=fc_output)
        return graph


class F_F_A_Behavior(Behavior):

    """Double layer feature conditions behavior"""

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)

        feature_condition_1 = ThresholdFeatureCondition(relation=">=", threshold=0)
        feature_condition_2 = ThresholdFeatureCondition(relation="<=", threshold=1)
        feature_condition_3 = ThresholdFeatureCondition(relation=">=", threshold=-1)

        graph.add_edge(feature_condition_1, feature_condition_2, index=True)
        graph.add_edge(feature_condition_1, feature_condition_3, index=False)

        for action, edge_index in zip(range(2, 4), (1, 0)):
            graph.add_edge(feature_condition_2, Action(action), index=edge_index)

        for action, edge_index in zip(range(2), (0, 1)):
            graph.add_edge(feature_condition_3, Action(action), index=edge_index)

        return graph


class F_AA_Behavior(Behavior):

    """Feature condition with mutliple actions on same index."""

    def __init__(self, name: str, any_mode: str) -> None:
        super().__init__(name, image=None)
        self.any_mode = any_mode

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self, any_mode=self.any_mode)
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)

        graph.add_edge(feature_condition, Action(0), index=int(True))
        graph.add_edge(feature_condition, Action(1), index=int(False))
        graph.add_edge(feature_condition, Action(2), index=int(False))

        return graph


class AF_A_Behavior(Behavior):

    """Double root with feature condition and action"""

    def __init__(self, name: str, any_mode: str) -> None:
        super().__init__(name, image=None)
        self.any_mode = any_mode

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self, any_mode=self.any_mode)

        graph.add_node(Action(0))
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)

        graph.add_edge(feature_condition, Action(1), index=int(True))
        graph.add_edge(feature_condition, Action(2), index=int(False))

        return graph



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Unit tests for the hebg.metaheuristics module. """

import numpy as np

import pytest_check as check

from hebg.layouts.metaheuristics import simulated_annealing


def test_simulated_annealing():
    """Simulated annealing must work on the simple x**2 case."""

    def energy(x):
        return x**2

    step_size = 0.05

    def neighbor(x):
        return x + np.random.choice([-1, 1]) * step_size

    optimal_x = simulated_annealing(
        -1, energy, neighbor, max_iterations=1000, initial_temperature=5, verbose=1
    )

    check.less_equal(abs(optimal_x), 3.1 * step_size)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=protected-access

""" Metaheuristics based layouts """

from copy import deepcopy

import networkx as nx
import numpy as np

from hebg.layouts.metaheuristics import simulated_annealing


def leveled_layout_energy(
    graph: nx.DiGraph, center=None, metaheuristic=simulated_annealing
):
    """Compute positions for a leveled DiGraph using a metaheuristic to minimize energy.

    Requires each node to have a 'level' attribute.

    Args:
        graph: A networkx DiGraph.
        center (Optional): Center of the graph layout.

    Returns:
        pos: Positions of each node.
        nodes_by_level: List of nodes by levels.

    """
    graph, center = nx.drawing.layout._process_params(graph, center, dim=2)

    nodes_by_level = graph.graph["nodes_by_level"]
    pos = {}
    step_size = 1 / max(len(nodes_by_level[level]) for level in nodes_by_level)
    spacing = np.arange(0, 1, step=step_size)
    for level in nodes_by_level:
        n_nodes_in_level = len(nodes_by_level[level])
        if n_nodes_in_level > 1:
            positions = np.linspace(
                0, len(spacing) - 1, n_nodes_in_level, endpoint=True, dtype=np.int32
            )
            positions = spacing[positions]
        else:
            positions = [spacing[(len(spacing) - 1) // 2]]

        for i, node in enumerate(nodes_by_level[level]):
            pos[node] = [level, positions[i]]

    def energy(pos, nodes_strenght=1, edges_strenght=2):
        def dist(start, stop):
            x_arr, y_arr = np.array(start), np.array(stop)
            return np.linalg.norm(x_arr - y_arr)

        energy = 0
        for level in nodes_by_level:
            for node in nodes_by_level[level]:
                energy += nodes_strenght * sum(
                    np.square(dist(pos[node], pos[n]))
                    for n in nodes_by_level[level]
                    if n != node
                )
                energy -= sum(
                    edges_strenght
                    / abs(
                        max(1, graph.nodes[node]["level"] - graph.nodes[pred]["level"])
                    )
                    / max(1e-6, dist(pos[node], pos[pred]))
                    for pred in graph.predecessors(node)
                )
                energy -= sum(
                    edges_strenght
                    / abs(
                        max(1, graph.nodes[node]["level"] - graph.nodes[succ]["level"])
                    )
                    / max(1e-6, dist(pos[node], pos[succ]))
                    for succ in graph.successors(node)
                )

        return energy

    def neighbor(pos: dict):
        pos_copy = deepcopy(pos)
        nodes_list = list(pos.keys())
        choosen_node_id = int(np.random.randint(len(nodes_list)))
        choosen_node = nodes_list[choosen_node_id]
        choosen_level = graph.nodes(data="level")[choosen_node]
        new_pos = [pos_copy[choosen_node][0], np.random.choice(spacing)]
        for node in nodes_by_level[choosen_level]:
            if node != choosen_node and np.all(np.isclose(new_pos, pos_copy[node])):
                pos_copy[choosen_node], pos_copy[node] = (
                    pos_copy[node],
                    pos_copy[choosen_node],
                )
                return pos_copy
        pos_copy[choosen_node] = new_pos
        return pos_copy

    return metaheuristic(pos, energy, neighbor)



from hebg.node import Action, EmptyNode
from hebg.behavior import Behavior
from hebg.heb_graph import HEBGraph

from tests.examples.feature_conditions import ThresholdFeatureCondition


class E_A_Behavior(Behavior):

    """Empty behavior"""

    def __init__(self, name: str, action: Action) -> None:
        super().__init__(name)
        self.action = action

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        graph.add_edge(EmptyNode("empty"), self.action)
        return graph


class E_F_A_Behavior(Behavior):

    """Double layer empty then feature conditions behavior"""

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        empty = EmptyNode("empty")
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)

        graph.add_edge(empty, feature_condition)
        for i, edge_index in zip(range(2), (0, 1)):
            action = Action(i)
            graph.add_edge(feature_condition, action, index=edge_index)

        return graph


class F_E_A_Behavior(Behavior):

    """Double layer feature conditions then empty behavior"""

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)

        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
        empty_0 = EmptyNode("empty_0")
        empty_1 = EmptyNode("empty_1")

        graph.add_edge(feature_condition, empty_0, index=int(True))
        graph.add_edge(feature_condition, empty_1, index=int(False))

        graph.add_edge(empty_0, Action(0))
        graph.add_edge(empty_1, Action(1))

        return graph


class E_E_A_Behavior(Behavior):

    """Double layer empty behavior"""

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)

        empty_0 = EmptyNode("empty_0")
        empty_1 = EmptyNode("empty_1")

        graph.add_edge(empty_0, empty_1)
        graph.add_edge(empty_1, Action(0))

        return graph



import pytest
import pytest_check as check

import networkx as nx
from hebg import HEBGraph
from hebg.unrolling import unroll_graph

from tests.examples.behaviors.loop import build_looping_behaviors

import matplotlib.pyplot as plt


class TestLoop:
    """Tests for the loop example"""

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.gather_wood, self.get_new_axe = build_looping_behaviors()

    def test_unroll_gather_wood(self):
        draw = False
        unrolled_graph = unroll_graph(self.gather_wood.graph)
        if draw:
            _plot_graph(unrolled_graph)

        expected_graph = nx.DiGraph()
        expected_graph.add_edge("Has axe", "Punch tree")
        expected_graph.add_edge("Has axe", "Cut tree with axe")
        expected_graph.add_edge("Has axe", "Has wood")

        # Expected sub-behavior
        expected_graph.add_edge("Has wood", "Gather wood")
        expected_graph.add_edge("Has wood", "Craft axe")
        check.is_true(nx.is_isomorphic(unrolled_graph, expected_graph))

    def test_unroll_get_new_axe(self):
        draw = False
        unrolled_graph = unroll_graph(self.get_new_axe.graph)
        if draw:
            _plot_graph(unrolled_graph)

        expected_graph = nx.DiGraph()
        expected_graph.add_edge("Has wood", "Has axe")
        expected_graph.add_edge("Has wood", "Craft new axe")

        # Expected sub-behavior
        expected_graph.add_edge("Has axe", "Punch tree")
        expected_graph.add_edge("Has axe", "Cut tree with axe")
        expected_graph.add_edge("Has axe", "Get new axe")
        check.is_true(nx.is_isomorphic(unrolled_graph, expected_graph))

    def test_unroll_gather_wood_cutting_alternatives(self):
        draw = False
        unrolled_graph = unroll_graph(
            self.gather_wood.graph,
            cut_looping_alternatives=True,
        )
        if draw:
            _plot_graph(unrolled_graph)

        expected_graph = nx.DiGraph()
        expected_graph.add_edge("Has axe", "Punch tree")
        expected_graph.add_edge("Has axe", "Cut tree with axe")
        expected_graph.add_edge("Has axe", "Has wood")

        # Expected sub-behavior
        expected_graph.add_edge("Has wood", "Punch tree")
        expected_graph.add_edge("Has wood", "Craft axe")
        check.is_true(nx.is_isomorphic(unrolled_graph, expected_graph))

    def test_unroll_get_new_axe_cutting_alternatives(self):
        draw = False
        unrolled_graph = unroll_graph(
            self.get_new_axe.graph,
            cut_looping_alternatives=True,
        )
        if draw:
            _plot_graph(unrolled_graph)

        expected_graph = nx.DiGraph()
        expected_graph.add_edge("Has wood", "Has axe")
        expected_graph.add_edge("Has wood", "Craft new axe")

        # Expected sub-behavior
        expected_graph.add_edge("Has axe", "Punch tree")
        expected_graph.add_edge("Has axe", "Cut tree with axe")
        check.is_true(nx.is_isomorphic(unrolled_graph, expected_graph))


def _plot_graph(graph: "HEBGraph"):
    _, ax = plt.subplots()
    graph.draw(ax)
    plt.show()



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=arguments-differ

""" Module containing the HEBGraph base class. """

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from matplotlib.axes import Axes
from networkx import DiGraph

from hebg.behavior import Behavior
from hebg.codegen import get_hebg_source
from hebg.draw import draw_hebgraph
from hebg.graph import get_roots, get_successors_with_index
from hebg.node import Node
from hebg.unrolling import unroll_graph


class HEBGraph(DiGraph):

    """Base class for Hierchical Explanation of Behavior as Graphs.

    An HEBGraph is a DiGraph, and as such stores nodes and directed edges with
    optional data, or attributes.

    But nodes of an HEBGraph are not arbitrary.
    Leaf nodes can either be an Action or a Behavior.
    Other nodes can either be a FeatureCondition or an EmptyNode.

    An HEBGraph determines a behavior, it can be called with an observation
    to return the action given by this behavior.

    An HEBGraph edges are directed and indexed,
    this indexing for path making when calling the graph.

    As in a DiGraph loops are allowed but multiple (parallel) edges are not.

    Args:
        behavior: The Behavior object from which this graph is built.
        all_behaviors: A dictionary of behavior, this can be used to avoid cirular definitions using
            the behavior names as anchor instead of the behavior object itself.
        any_mode: How to choose path, when multiple path are valid.
        incoming_graph_data: Additional data to include in the graph.

    """

    NODES_COLORS = {"feature_condition": "blue", "action": "red", "behavior": "orange"}
    EDGES_COLORS = {
        0: "red",
        1: "green",
        2: "blue",
        3: "yellow",
        4: "purple",
        5: "cyan",
        6: "gray",
    }
    ANY_MODES = ("first", "last", "random")

    def __init__(
        self,
        behavior: Behavior,
        all_behaviors: Dict[str, Behavior] = None,
        incoming_graph_data=None,
        any_mode: str = "first",
        **attr,
    ):
        self.behavior = behavior
        self.all_behaviors = all_behaviors if all_behaviors is not None else {}

        self._unrolled_graph = None

        assert any_mode in self.ANY_MODES, f"Unknowed any_mode: {any_mode}"
        self.any_mode = any_mode

        super().__init__(incoming_graph_data=incoming_graph_data, **attr)

    def add_node(self, node_for_adding: Node, **attr):
        node = node_for_adding
        color = attr.pop("color", None)
        attr.pop("type", None)
        attr.pop("image", None)
        if color is None:
            try:
                color = self.NODES_COLORS[node.type]
            except KeyError:
                color = None
        super().add_node(node, type=node.type, color=color, image=node.image, **attr)

    def add_edge(self, u_of_edge: Node, v_of_edge: Node, index: int = 1, **attr):
        for node in (u_of_edge, v_of_edge):
            if node not in self.nodes():
                self.add_node(node)

        color = attr.pop("color", None)
        if color is None:
            try:
                color = self.EDGES_COLORS[index]
            except KeyError:
                color = "black"
        super().add_edge(u_of_edge, v_of_edge, index=index, color=color, **attr)

    def _get_any_action(
        self, nodes: List[Node], observation, behaviors_in_search: list
    ):
        actions = []
        for node in nodes:
            action = self._get_action(node, observation, behaviors_in_search)
            if action is None:
                return None
            actions.append(action)
        actions = [action for action in actions if action != "Impossible"]
        if len(actions) == 0:
            return "Impossible"
        if self.any_mode == "first":
            return actions[0]
        if self.any_mode == "last":
            return actions[-1]
        if self.any_mode == "random":
            return np.random.choice(actions)

    @property
    def unrolled_graph(self) -> HEBGraph:
        """Access to the unrolled behavior graph.

        The unrolled behavior graph as the same behavior but every behavior node is recursively replaced
        by it's behavior graph if it can be computed.

        Only build's the graph the first time called for efficiency.

        Returns:
            This HEBGraph's unrolled HEBGraph.

        """
        if self._unrolled_graph is None:
            self._unrolled_graph = unroll_graph(self)
        return self._unrolled_graph

    def _get_action(self, node: Node, observation: Any, behaviors_in_search: List[str]):
        # Behavior
        if node.type == "behavior":

            # To avoid cycling definitions
            if node.name in behaviors_in_search:
                return "Impossible"

            # Search for name reference in all_behaviors
            if node.name in self.all_behaviors:
                node = self.all_behaviors[node.name]

            return node(observation, behaviors_in_search)

        # Action
        if node.type == "action":
            return node(observation)
        # Feature Condition
        if node.type == "feature_condition":
            next_edge_index = int(node(observation))
            next_nodes = get_successors_with_index(self, node, next_edge_index)
            return self._get_any_action(next_nodes, observation, behaviors_in_search)
        # Empty
        if node.type == "empty":
            next_node = self.successors(node).__next__()
            return self._get_action(next_node, observation, behaviors_in_search)
        raise ValueError(f"Unknowed value {node.type} for node.type with node: {node}.")

    def __call__(
        self,
        observation,
        behaviors_in_search: Optional[List[str]] = None,
    ) -> Any:
        if behaviors_in_search is None:
            behaviors_in_search = []
        behaviors_in_search.append(self.behavior.name)
        return self._get_any_action(self.roots, observation, behaviors_in_search)

    @property
    def roots(self) -> List[Node]:
        """Roots of the behavior graph (nodes without predecessors)."""
        return get_roots(self)

    def generate_source_code(self) -> str:
        """Generated source code of the behavior from graph."""
        return get_hebg_source(self)

    def draw(
        self, ax: "Axes", **kwargs
    ) -> Tuple["Axes", Dict[Node, Tuple[float, float]]]:
        """Draw the HEBGraph on the given Axis.

        Args:
            ax: The matplotlib ax to draw on.

        Kwargs:
            fontcolor: Font color to use for all texts.

        Returns:
            The resulting matplotlib Axis drawn on and a dictionary of each node position.

        """
        return draw_hebgraph(self, ax, **kwargs)



from typing import Union
from enum import Enum

from hebg.node import FeatureCondition


class ThresholdFeatureCondition(FeatureCondition):
    class Relation(Enum):
        GREATER_OR_EQUAL_TO = ">="
        LESSER_OR_EQUAL_TO = "<="
        GREATER_THAN = ">"
        LESSER_THAN = "<"

    def __init__(
        self, relation: Union[Relation, str] = ">=", threshold: float = 0
    ) -> None:
        """Threshold-based feature condition for scalar feature."""
        self.relation = relation
        self.threshold = threshold
        self._relation = self.Relation(relation)
        display_name = self._relation.name.capitalize().replace("_", " ")
        name = f"{display_name} {threshold} ?"
        super().__init__(name=name, image=None)

    def __call__(self, observation: float) -> int:
        conditions = {
            self.Relation.GREATER_OR_EQUAL_TO: int(observation >= self.threshold),
            self.Relation.LESSER_OR_EQUAL_TO: int(observation <= self.threshold),
            self.Relation.GREATER_THAN: int(observation > self.threshold),
            self.Relation.LESSER_THAN: int(observation < self.threshold),
        }
        if self._relation in conditions:
            return conditions[self._relation]


class IsDivisibleFeatureCondition(FeatureCondition):
    def __init__(self, number: int = 0) -> None:
        """Is divisible feature condition for scalar feature."""
        self.number = number
        name = f"Is divisible by {number} ?"
        super().__init__(name=name, image=None)

    def __call__(self, observation: float) -> int:
        return int(observation // self.number == 1)



from typing import List

from hebg import HEBGraph, Action, FeatureCondition, Behavior


class GatherWood(Behavior):
    """Gather wood"""

    def __init__(self) -> None:
        """Gather wood"""
        super().__init__("Gather wood")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        feature = FeatureCondition("Has an axe")
        graph.add_edge(feature, Action("Punch tree"), index=False)
        graph.add_edge(feature, Behavior("Get new axe"), index=False)
        graph.add_edge(feature, Action("Use axe on tree"), index=True)
        return graph


class GetNewAxe(Behavior):
    """Get new axe with wood"""

    def __init__(self) -> None:
        """Get new axe with wood"""
        super().__init__("Get new axe")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        feature = FeatureCondition("Has wood")
        graph.add_edge(feature, Behavior("Gather wood"), index=False)
        graph.add_edge(feature, Action("Craft axe"), index=True)
        return graph


def build_looping_behaviors() -> List[Behavior]:
    behaviors: List[Behavior] = [GatherWood(), GetNewAxe()]
    all_behaviors = {behavior.name: behavior for behavior in behaviors}
    for behavior in behaviors:
        behavior.graph.all_behaviors = all_behaviors
    return behaviors



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=arguments-differ

""" Module for building underlying requirement graphs based on a set of behaviors. """

from __future__ import annotations

from copy import deepcopy
from typing import Dict, List

from networkx import DiGraph, descendants

from hebg.behavior import Behavior
from hebg.graph import compute_levels
from hebg.heb_graph import HEBGraph
from hebg.node import EmptyNode


def build_requirement_graph(behaviors: List[Behavior]) -> DiGraph:
    """Builds a DiGraph of the requirements induced by a list of behaviors.

    Args:
        behaviors: List of Behaviors to build the requirement graph from.

    Returns:
        The requirement graph induced by the given list of behaviors.

    """

    try:
        heb_graphs = [behavior.graph for behavior in behaviors]
    except NotImplementedError as error:
        user_msg = "All behaviors given must be able to build an HEBGraph"
        raise NotImplementedError(user_msg) from error

    requirements_graph = DiGraph()
    for behavior in behaviors:
        requirements_graph.add_node(behavior)

    requirement_degree = {}

    for graph in heb_graphs:
        requirement_degree[graph.behavior] = {}
        for node in graph.nodes():
            if not isinstance(node, EmptyNode):
                continue
            requirement_degree = _cut_alternatives_to_empty_node(
                graph, node, requirement_degree
            )

    for graph in heb_graphs:
        for node in graph.nodes():
            if not isinstance(node, Behavior):
                continue
            if node not in requirement_degree[graph.behavior]:
                requirement_degree[graph.behavior][node] = 0
            requirement_degree[graph.behavior][node] += 1

    index = 0
    for graph in heb_graphs:
        for node in graph.nodes():
            if (
                not isinstance(node, Behavior)
                or requirement_degree[graph.behavior][node] == 0
            ):
                continue
            if node not in requirements_graph.nodes():
                requirements_graph.add_node(node)
            index = len(list(requirements_graph.successors(node))) + 1
            requirements_graph.add_edge(node, graph.behavior, index=index)

    compute_levels(requirements_graph)
    return requirements_graph


def _cut_alternatives_to_empty_node(
    graph: HEBGraph,
    node: EmptyNode,
    requirement_degree: Dict[Behavior, Dict[Behavior, int]],
) -> Dict[Behavior, Dict[Behavior, int]]:
    successor = list(graph.successors(node))[0]
    empty_index = graph.edges[node, successor]["index"]
    alternatives = graph.predecessors(successor)
    alternatives = [
        alt_node
        for alt_node in alternatives
        if graph.edges[alt_node, successor]["index"] == empty_index
    ]
    cut_graph = deepcopy(graph)
    for alternative in alternatives:
        cut_graph.remove_edge(alternative, successor)
    for alternative in alternatives:
        following_behaviors = [
            following_node
            for following_node in descendants(cut_graph, alternative)
            if isinstance(following_node, Behavior)
        ]
        for following_behavior in following_behaviors:
            if following_behavior not in requirement_degree[graph.behavior]:
                requirement_degree[graph.behavior][following_behavior] = 0
            requirement_degree[graph.behavior][following_behavior] -= 1

    return requirement_degree



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" A structure for explainable hierarchical reinforcement learning """

from hebg.behavior import Behavior
from hebg.heb_graph import HEBGraph
from hebg.node import Action, EmptyNode, FeatureCondition, Node, StochasticAction
from hebg.requirements_graph import build_requirement_graph



from hebg import HEBGraph, Action, FeatureCondition, Behavior


class Behavior0(Behavior):
    """Behavior 0"""

    def __init__(self) -> None:
        super().__init__("behavior 0")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        feature = FeatureCondition("feature 0", complexity=1)
        graph.add_edge(feature, Action(0, complexity=1), index=False)
        graph.add_edge(feature, Action(1, complexity=1), index=True)
        return graph


class Behavior1(Behavior):
    """Behavior 1"""

    def __init__(self) -> None:
        super().__init__("behavior 1")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        feature_1 = FeatureCondition("feature 1", complexity=1)
        feature_2 = FeatureCondition("feature 2", complexity=1)
        graph.add_edge(feature_1, Behavior0(), index=False)
        graph.add_edge(feature_1, feature_2, index=True)
        graph.add_edge(feature_2, Action(0, complexity=1), index=False)
        graph.add_edge(feature_2, Action(2, complexity=1), index=True)
        return graph


class Behavior2(Behavior):
    """Behavior 2"""

    def __init__(self) -> None:
        super().__init__("behavior 2")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        feature_3 = FeatureCondition("feature 3", complexity=1)
        feature_4 = FeatureCondition("feature 4", complexity=1)
        feature_5 = FeatureCondition("feature 5", complexity=1)
        graph.add_edge(feature_3, feature_4, index=False)
        graph.add_edge(feature_3, feature_5, index=True)
        graph.add_edge(feature_4, Action(0, complexity=1), index=False)
        graph.add_edge(feature_4, Behavior1(), index=True)
        graph.add_edge(feature_5, Behavior1(), index=False)
        graph.add_edge(feature_5, Behavior0(), index=True)
        return graph



from hebg import Action, Behavior
from tests.examples.behaviors import F_A_Behavior
from tests.examples.feature_conditions import IsDivisibleFeatureCondition


def build_binary_sum_behavior() -> F_A_Behavior:
    feature_condition = IsDivisibleFeatureCondition(2)
    actions = {0: Action(0), 1: Action(1)}
    binary_1 = F_A_Behavior("Is x1 in binary ?", feature_condition, actions)

    feature_condition = IsDivisibleFeatureCondition(2)
    actions = {0: Action(1), 1: Action(0)}
    binary_0 = F_A_Behavior("Is x0 in binary ?", feature_condition, actions)

    feature_condition = IsDivisibleFeatureCondition(4)
    actions = {0: Action(0), 1: binary_1}
    binary_11 = F_A_Behavior("Is x11 in binary ?", feature_condition, actions)

    feature_condition = IsDivisibleFeatureCondition(4)
    actions = {0: binary_0, 1: binary_1}
    binary_10_01 = F_A_Behavior("Is x01 or x10 in binary ?", feature_condition, actions)

    feature_condition = IsDivisibleFeatureCondition(8)
    actions = {0: binary_11, 1: binary_10_01}

    return F_A_Behavior("Is sum (of last 3 binary) 2 ?", feature_condition, actions)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" General complexity. """


from typing import TYPE_CHECKING, Dict, Tuple

from hebg.behavior import Behavior
from hebg.metrics.complexity.utils import update_sum_dict
from hebg.node import Action

if TYPE_CHECKING:
    from hebg.node import Node


def general_complexity(
    behavior: Behavior,
    used_nodes_all: Dict["Node", Dict["Node", int]],
    scomplexity,
    kcomplexity,
    previous_used_nodes: Dict["Node", int] = None,
) -> Tuple[float]:
    """Compute the general complexity of a Behavior with used nodes.

    Using the number of time each node is used in its HEBGraph and based on the increase of
    complexity given by 'kcomplexity', and the saved complexity using behaviors given by
    'saved_complexity', we sum the general complexity of an behavior and the total saved complexity.

    Args:
        behavior: The Behavior for which we compute the general complexity.
        used_nodes_all: Dictionary of dictionary of the number of times each nodes was used in the
            past, and thus for each node. Not being in a dictionary is counted as not being used.
        scomplexity: Callable taking as input (node, n_used, n_previous_used), where node is
            the node concerned, n_used is the number of time this node is used and n_previous_used
            is the number of time this node was used before that, returning the saved complexity
            thanks to using a behavior.
        kcomplexity: Callable taking as input (node, n_used), where node is the node concerned and
            n_used is the number of time this node is used, returning the accumulated complexity.
        previous_used_nodes: Dictionary of the number of times each nodes was used in the past,
            not being in the dictionary is counted as 0.

    Returns:
        Tuple composed of the general complexity and the total saved complexity.

    """

    previous_used_nodes = previous_used_nodes if previous_used_nodes else {}

    total_complexity = 0
    saved_complexity = 0

    for node, n_used in used_nodes_all[behavior].items():
        n_previous_used = (
            previous_used_nodes[node] if node in previous_used_nodes else 0
        )

        if isinstance(node, Behavior) and node in used_nodes_all:
            node_complexity, saved_node_complexity = general_complexity(
                node,
                used_nodes_all,
                scomplexity=scomplexity,
                kcomplexity=kcomplexity,
                previous_used_nodes=previous_used_nodes.copy(),
            )
            previous_used_nodes = update_sum_dict(
                previous_used_nodes, used_nodes_all[node]
            )
            total_complexity += saved_node_complexity * kcomplexity(node, n_used)
            saved_complexity += saved_node_complexity * kcomplexity(node, n_used)
        else:
            node_complexity = node.complexity

        total_complexity += node_complexity * kcomplexity(node, n_used)

        if isinstance(node, (Behavior, Action)):
            saved_complexity += node_complexity * scomplexity(
                node, n_used, n_previous_used
            )

        previous_used_nodes = update_sum_dict(previous_used_nodes, {node: n_used})

    return total_complexity - saved_complexity, saved_complexity


def learning_complexity(
    behavior: Behavior,
    used_nodes_all: Dict["Node", Dict["Node", int]],
    previous_used_nodes=None,
):
    """Compute the learning complexity of a Behavior with used nodes.

    Using the number of time each node is used in its HEBGraph we compute the learning
    complexity of a behavior and the total saved complexity.

    Args:
        behavior: The Behavior for which we compute the learning complexity.
        used_nodes_all: Dictionary of dictionary of the number of times each nodes was used in the
            past, and thus for each node. Not being in a dictionary is counted as not being used.
        previous_used_nodes: Dictionary of the number of times each nodes was used in the past,
            not being in the dictionary is counted as not being used.

    Returns:
        Tuple composed of the learning complexity and the total saved complexity.

    """
    return general_complexity(
        behavior=behavior,
        used_nodes_all=used_nodes_all,
        previous_used_nodes=previous_used_nodes,
        scomplexity=lambda node, k, p: max(0, min(k, p + k - 1)),
        kcomplexity=lambda node, k: k,
    )



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Module for base Behavior. """

from __future__ import annotations

from typing import TYPE_CHECKING

from hebg.graph import compute_levels
from hebg.node import Node

if TYPE_CHECKING:
    from hebg.heb_graph import HEBGraph


class Behavior(Node):

    """Abstract class for a Behavior as Node"""

    def __init__(self, name: str, image=None) -> None:
        super().__init__(name, "behavior", image=image)
        self._graph = None

    def __call__(self, observation, *args, **kwargs):
        """Use the behavior to get next actions.

        By default, uses the HEBGraph if it can be built.

        Args:
            observation: Observations of the environment.
            greedy: If true, the agent should act greedily.

        Returns:
            action: Action given by the behavior with current observation.

        """
        return self.graph.__call__(observation, *args, **kwargs)

    def build_graph(self) -> HEBGraph:
        """Build the HEBGraph of this Behavior.

        Returns:
            The built HEBGraph.

        """
        raise NotImplementedError

    @property
    def graph(self) -> HEBGraph:
        """Access to the Behavior's graph.

        Only build's the graph the first time called for efficiency.

        Returns:
            This Behavior's HEBGraph.

        """
        if self._graph is None:
            self._graph = self.build_graph()
            compute_levels(self._graph)
        return self._graph



"""This examples shows how could we hierarchicaly build a behavior to pet a cat.

Here is the hierarchical structure that we would want:

```
PetACat:
    IsThereACatAround ?
    -> Yes:
        PetNearbyCat
    -> No:
        LookForACat

PetNearbyCat:
    IsYourHandNearTheCat ?
    -> Yes:
        Pet
    -> No:
        MoveYourHandNearTheCat
```

"""

import pytest
import pytest_check as check

import matplotlib.pyplot as plt

from hebg import HEBGraph, Action, FeatureCondition, Behavior
from hebg.unrolling import unroll_graph
from tests.integration.test_code_generation import _unidiff_output


class Pet(Action):
    def __init__(self) -> None:
        super().__init__(action="Pet")


class IsYourHandNearTheCat(FeatureCondition):
    def __init__(self, hand) -> None:
        super().__init__(name="Is hand near the cat ?")
        self.hand = hand

    def __call__(self, observation):
        # Could be a very complex function that returns 1 is the hand is near the cat else 0.
        if observation["cat"] == observation[self.hand]:
            return int(True)  # 1
        return int(False)  # 0


class MoveYourHandNearTheCat(Behavior):
    def __init__(self) -> None:
        super().__init__(name="Move slowly your hand near the cat")

    def __call__(self, observation, *args, **kwargs) -> Action:
        # Could be a very complex function that returns actions from any given observation
        return Action("Move hand to cat")


class PetNearbyCat(Behavior):
    def __init__(self) -> None:
        super().__init__(name="Pet nearby cat")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        is_hand_near_cat = IsYourHandNearTheCat(hand="hand")
        graph.add_edge(is_hand_near_cat, MoveYourHandNearTheCat(), index=int(False))
        graph.add_edge(is_hand_near_cat, Pet(), index=int(True))

        return graph


class IsThereACatAround(FeatureCondition):
    def __init__(self) -> None:
        super().__init__(name="Is there a cat around ?")

    def __call__(self, observation):
        # Could be a very complex function that returns 1 is there is a cat around else 0.
        if "cat" in observation:
            return int(True)  # 1
        return int(False)  # 0


class LookForACat(Behavior):
    def __init__(self) -> None:
        super().__init__(name="Look for a nearby cat")

    def __call__(self, observation, *args, **kwargs) -> Action:
        # Could be a very complex function that returns actions from any given observation
        return Action("Move to a cat")


class PetACat(Behavior):
    def __init__(self) -> None:
        super().__init__(name="Pet a cat")

    def build_graph(self) -> HEBGraph:
        graph = HEBGraph(self)
        is_a_cat_around = IsThereACatAround()
        graph.add_edge(is_a_cat_around, LookForACat(), index=int(False))
        graph.add_edge(is_a_cat_around, PetNearbyCat(), index=int(True))
        return graph


class TestPetACat:
    """PetACat example"""

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.pet_nearby_cat_behavior = PetNearbyCat()
        self.pet_a_cat_behavior = PetACat()

    def test_call(self):
        """should give expected call"""
        observation = {
            "cat": "sofa",
            "hand": "computer",
        }

        # Call on observation
        action = self.pet_a_cat_behavior(observation)
        check.equal(action, Action("Move hand to cat"))

    def test_pet_a_cat_graph_edges(self):
        """should give expected edges for PetACat"""
        # Obtain networkx graph
        graph = self.pet_a_cat_behavior.graph
        check.equal(
            set(graph.edges(data="index")),
            {
                ("Is there a cat around ?", "Look for a nearby cat", 0),
                ("Is there a cat around ?", "Pet nearby cat", 1),
            },
        )

    def test_pet_nearby_cat_graph_edges(self):
        """should give expected edges for PetNearbyCat"""
        # Obtain networkx graph
        graph = self.pet_nearby_cat_behavior.graph
        check.equal(
            set(graph.edges(data="index")),
            {
                ("Is hand near the cat ?", "Move slowly your hand near the cat", 0),
                ("Is hand near the cat ?", "Action(Pet)", 1),
            },
        )

    def test_draw(self):
        """should be able to draw without error"""
        fig, ax = plt.subplots()
        self.pet_a_cat_behavior.graph.draw(ax)
        plt.close(fig)

    def test_draw_unrolled(self):
        """should be able to draw without error"""
        fig, ax = plt.subplots()
        unrolled_graph = unroll_graph(self.pet_a_cat_behavior.graph)
        unrolled_graph.draw(ax)
        plt.close(fig)

    @pytest.mark.filterwarnings("ignore:Could not load graph for behavior")
    def test_codegen(self):
        """should generate expected source code"""
        code = self.pet_a_cat_behavior.graph.generate_source_code()
        expected_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "# Require 'Look for a nearby cat' behavior to be given.",
                "# Require 'Move slowly your hand near the cat' behavior to be given.",
                "class PetACat(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Is there a cat around ?'](observation)",
                "        if edge_index == 0:",
                "            return self.known_behaviors['Look for a nearby cat'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['Is hand near the cat ?'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.known_behaviors['Move slowly your hand near the cat'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['Action(Pet)'](observation)",
            )
        )
        check.equal(code, expected_code, _unidiff_output(code, expected_code))
